#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

from selinon_test_case import SelinonTestCase
from selinon import SystemState


class TestSystemState(SelinonTestCase):
    def test_simple1(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     Task2
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Run without change at first
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert system_state.node_args is None
        assert retry is not None
        assert 'Task1' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, 1)

        # check dispatcher_id propagation
        assert task1.dispatcher_id == id(self)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert len(state_dict['active_nodes']) == 1
        assert len(state_dict['finished_nodes']) == 1
        assert len(state_dict['waiting_edges']) == 1
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' not in state_dict['finished_nodes'].keys()
        # the result of Task1 should be propagated to Task2
        assert system_state.node_args is None

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2, 1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        # All tasks done
        assert retry is None
        assert len(state_dict['active_nodes']) == 0
        assert len(state_dict['finished_nodes']) == 2
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()

    def test_2_to_1(self):
        #
        # flow1:
        #
        #    Task1      Task2
        #      |          |
        #       ----------
        #           |
        #         Task3
        #
        edge_table = {
                      'flow1': [{'from': ['Task1', 'Task2'], 'to': ['Task3'], 'condition': self.cond_true},
                                {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Run without change at first
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Task2 has finished
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert len(state_dict.get('active_nodes')) == 1
        assert len(state_dict.get('finished_nodes')) == 1
        assert state_dict['active_nodes'][0]['name'] == 'Task1'
        assert 'Task2' in state_dict['finished_nodes'].keys()
        assert 'Task1' not in state_dict['finished_nodes'].keys()
        assert 'Task3' not in state_dict['finished_nodes'].keys()
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Task1 has finished
        self.set_finished(task1, "some result of task1 that shouldn't be propagated")

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('active_nodes')) == 1
        assert len(state_dict.get('finished_nodes')) == 2
        assert state_dict['active_nodes'][0]['name'] == 'Task3'
        assert 'Task2' in state_dict['finished_nodes'].keys()
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task3' not in state_dict['finished_nodes'].keys()
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # No change so far
        task3 = self.get_task('Task3')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('active_nodes')) == 1
        assert len(state_dict.get('finished_nodes')) == 2
        assert state_dict['active_nodes'][0]['name'] == 'Task3'
        assert 'Task2' in state_dict['finished_nodes'].keys()
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task3' not in state_dict['finished_nodes'].keys()
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Task 3 has finished
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert len(state_dict.get('active_nodes')) == 0
        assert len(state_dict.get('finished_nodes')) == 3
        assert 'Task2' in state_dict['finished_nodes'].keys()
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task3' in state_dict['finished_nodes'].keys()

    def test_2_to_1_separate(self):
        #
        # flow1:
        #
        #    Task1   Task2
        #      \      /
        #       \    /
        #        \  /
        #        Task3
        #
        # Note:
        #   Each time only one task instance finishes
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']

        # Run without change at first
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']

        # Task1 has finished
        self.set_finished(task1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('active_nodes')) == 2
        assert state_dict['active_nodes'][0]['name'], ['Task2' in 'Task3']
        assert state_dict['active_nodes'][1]['name'], ['Task2' in 'Task3']
        assert 'Task1' in state_dict['finished_nodes'].keys()

        # No change so far
        task3_0 = self.get_task('Task3')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('active_nodes')) == 2
        assert state_dict['active_nodes'][0]['name'], ['Task2' in 'Task3']
        assert state_dict['active_nodes'][1]['name'], ['Task2' in 'Task3']
        assert 'Task1' in state_dict['finished_nodes'].keys()

        # Task2 has finished
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('active_nodes')) == 2
        assert state_dict['active_nodes'][0]['name'] == 'Task3'
        assert state_dict['active_nodes'][1]['name'] == 'Task3'
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()

        # Again, no change
        task3_1 = self.get_task('Task3')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('active_nodes')) == 2
        assert state_dict['active_nodes'][0]['name'] == 'Task3'
        assert state_dict['active_nodes'][1]['name'] == 'Task3'
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()

        # Now the second instance of Task3 has finished
        self.set_finished(task3_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('active_nodes')) == 1
        assert state_dict['active_nodes'][0]['name'] == 'Task3'
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()
        assert 'Task3' in state_dict['finished_nodes'].keys()

        # The last task - the instance of Task3 has just finished
        self.set_finished(task3_0)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert len(state_dict.get('active_nodes')) == 0
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict['finished_nodes']['Task3']) == 2
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()

    def test_2_to_1_separate_2(self):
        #
        # flow1:
        #
        #    Task1   Task2
        #      \      /
        #       \    /
        #        \  /
        #        Task3
        #
        # Note:
        #   Task1 and Task2 finish at the same time
        #   Both instances of Task3 finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task2'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']

        # Task1, Task2 has finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        self.set_finished(task1)
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('active_nodes')) == 2
        assert state_dict['active_nodes'][0]['name'] == 'Task3'
        assert state_dict['active_nodes'][1]['name'] == 'Task3'
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()

        # Now both Task3 instances finished
        task3_0 = self.get_task('Task3', 0)
        task3_1 = self.get_task('Task3', 1)
        self.set_finished(task3_0)
        self.set_finished(task3_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert len(state_dict.get('active_nodes')) == 0
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict['finished_nodes']['Task3']) == 2
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()

    def test_3_to_1(self):
        #
        # flow1:
        #
        #    Task1    Task2   Task3
        #      |        |       |
        #       ----------------
        #               |
        #             Task4
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2', 'Task3'], 'to': ['Task4'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2', 'Task3'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']

        # Task1 has finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        self.set_finished(task1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('active_nodes')) == 2
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task4' not in self.instantiated_tasks

        # Task2 and Task3 have finished
        self.set_finished(task2)
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('active_nodes')) == 1
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()
        assert 'Task3' in state_dict['finished_nodes'].keys()
        assert 'Task4' in self.instantiated_tasks

        # Wait for Task4
        task4 = self.get_task('Task4')

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('active_nodes')) == 1
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()
        assert 'Task3' in state_dict['finished_nodes'].keys()
        assert 'Task4' in self.instantiated_tasks

        # Finish the flow
        self.set_finished(task4)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert len(state_dict.get('active_nodes')) == 0
        assert len(state_dict.get('finished_nodes')) == 4
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()
        assert 'Task3' in state_dict['finished_nodes'].keys()
        assert 'Task4' in state_dict['finished_nodes'].keys()

    def test_3_to_1_fail(self):
        #
        # flow1:
        #
        #    Task1    Task2   Task3
        #      |        |       |
        #       ----------------
        #               |
        #               X
        #               |
        #             Task4
        #
        # Note:
        #   The condition will fail
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2', 'Task3'], 'to': ['Task4'], 'condition': self.cond_false},
                      {'from': [], 'to': ['Task1', 'Task2', 'Task3'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']

        # Task1, Task2 and Task3 have finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        self.set_finished(task1)
        self.set_finished(task2)
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert len(state_dict.get('active_nodes')) == 0
        assert len(state_dict.get('finished_nodes')) == 3
        assert 'Task1' in state_dict['finished_nodes'].keys()
        assert 'Task2' in state_dict['finished_nodes'].keys()
        assert 'Task3' in state_dict['finished_nodes'].keys()
        assert 'Task4' not in state_dict['finished_nodes'].keys()

    def test_1_to_2_separate(self):
        #
        # flow1:
        #
        #         Task1
        #           |
        #       ----------
        #      |          |
        #    Task2      Task3
        #
        # Note:
        #   Task3 finishes before Task2
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 1

        # No change first
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 1

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, 0xDEADBEEF)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 2

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 2

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 2
        assert len(state_dict.get('active_nodes')) == 1

        # No change so far
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 2
        assert len(state_dict.get('active_nodes')) == 1

        # Task2 has finished
        task2= self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict.get('active_nodes')) == 0

    def test_1_to_2(self):
        #
        # flow1:
        #
        #         Task1
        #           |
        #       ----------
        #      |          |
        #    Task2      Task3
        #
        # Note:
        #   Task2 and Task3 finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 1

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, 0xDEADBEEF)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 2

        # Task2, Task3 have finished
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        self.set_finished(task2)
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict.get('active_nodes')) == 0

    def test_1_to_3_separate(self):
        #
        # flow1:
        #
        #             Task1
        #               |
        #       ----------------
        #      |        |       |
        #    Task2    Task3   Task4
        #
        # Note:
        #   Child tasks do not finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3', 'Task4'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 1

        # Task1 has finished
        task1_result = {'foo': ['bar']}
        task1 = self.get_task('Task1')
        self.set_finished(task1, task1_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 3

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 2
        assert len(state_dict.get('active_nodes')) == 2

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict.get('active_nodes')) == 1

        # Task4 has finished
        task4 = self.get_task('Task4')
        self.set_finished(task4)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 4
        assert len(state_dict.get('active_nodes')) == 0

    def test_1_to_3(self):
        #
        # flow1:
        #
        #             Task1
        #               |
        #       ----------------
        #      |        |       |
        #    Task2    Task3   Task4
        #
        # Note:
        #   Child tasks finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3', 'Task4'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 1

        # Task1 has finished
        task1_result = {'foo': ['bar']}
        task1 = self.get_task('Task1')
        self.set_finished(task1, task1_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 3

        # Task2 has finished
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        task4 = self.get_task('Task4')
        self.set_finished(task2)
        self.set_finished(task3)
        self.set_finished(task4)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 4
        assert len(state_dict.get('active_nodes')) == 0

    def test_3_to_3(self):
        #
        # flow1:
        #
        #    Task1    Task2   Task3
        #      |        |       |
        #       ----------------
        #               |
        #               |
        #       ----------------
        #      |        |       |
        #    Task4    Task5   Task6
        #
        # Note:
        #   Task1 and Task2 finish at the same time, Task3 afterwards. Task4, Task5 and Task6 finish at the same time
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2', 'Task3'], 'to': ['Task4', 'Task5', 'Task6'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2', 'Task3'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' not in self.instantiated_tasks
        assert 'Task5' not in self.instantiated_tasks
        assert 'Task6' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 3

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' not in self.instantiated_tasks
        assert 'Task5' not in self.instantiated_tasks
        assert 'Task6' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 3

        # Task1, Task2 have finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        self.set_finished(task1)
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' not in self.instantiated_tasks
        assert 'Task5' not in self.instantiated_tasks
        assert 'Task6' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 2
        assert len(state_dict.get('active_nodes')) == 1

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' not in self.instantiated_tasks
        assert 'Task5' not in self.instantiated_tasks
        assert 'Task6' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 2
        assert len(state_dict.get('active_nodes')) == 1

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert 'Task5' in self.instantiated_tasks
        assert 'Task6' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict.get('active_nodes')) == 3

        # Task4, Task5, Task6 have finished
        task4 = self.get_task('Task4')
        task5 = self.get_task('Task5')
        task6 = self.get_task('Task6')
        self.set_finished(task4)
        self.set_finished(task5)
        self.set_finished(task6)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert 'Task5' in self.instantiated_tasks
        assert 'Task6' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 6
        assert len(state_dict.get('active_nodes')) == 0

    def test_3_to_3_2(self):
        #
        # flow1:
        #
        #    Task1    Task2   Task3
        #      |        |       |
        #       ----------------
        #               |
        #               |
        #       ----------------
        #      |        |       |
        #    Task4    Task5   Task6
        #
        # Note:
        #   Task1, Task2 and Task3 finish at the same time; Task4, Task5 finish at the same time, then Task6
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2', 'Task3'], 'to': ['Task4', 'Task5', 'Task6'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2', 'Task3'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' not in self.instantiated_tasks
        assert 'Task5' not in self.instantiated_tasks
        assert 'Task6' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 3

        # Task1, Task2, Task3 have finished
        task1 = self.get_task('Task1')
        task2 = self.get_task('Task2')
        task3 = self.get_task('Task3')
        self.set_finished(task1)
        self.set_finished(task2)
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert 'Task5' in self.instantiated_tasks
        assert 'Task6' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict.get('active_nodes')) == 3

        # Task4 has finished
        task4 = self.get_task('Task4')
        self.set_finished(task4)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert 'Task5' in self.instantiated_tasks
        assert 'Task6' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 4
        assert len(state_dict.get('active_nodes')) == 2

        # Task5, Task6 have finished
        task5 = self.get_task('Task5')
        task6 = self.get_task('Task6')
        self.set_finished(task5)
        self.set_finished(task6)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert 'Task4' in self.instantiated_tasks
        assert 'Task5' in self.instantiated_tasks
        assert 'Task6' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 6
        assert len(state_dict.get('active_nodes')) == 0

    def test_1_to_2_recursive(self):
        #
        # flow1:
        #
        #         Task1  <------
        #           |           |
        #       ----------      |
        #      |          |     |
        #    Task2      Task3 --
        #
        # Note:
        #   Task2 finishes before Task3 in the first iteration
        #   Task3 finishes before Task2 in the second iteration
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['Task2', 'Task3'], 'condition': self.cond_true},
                      {'from': ['Task3'], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        node_args = "some task arguments"
        system_state = SystemState(id(self), 'flow1', node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 1

        # No change so far
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' not in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 1

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1, "some task result")

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 2

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 2

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 2
        assert len(state_dict.get('active_nodes')) == 1

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 2
        assert len(state_dict.get('active_nodes')) == 1

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(self.get_task_instance.task_by_name('Task1')) == 2
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict.get('active_nodes')) == 1

        # The next iteration
        # No change first
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(self.get_task_instance.task_by_name('Task1')) == 2
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict.get('active_nodes')) == 1

        # The second instance of Task1 has finished
        task1_1 = self.get_task('Task1')
        self.set_finished(task1_1)
        # We do not set task result, since it should be propagated from the first run of Task1

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        # There should be a second instance of Task2 and Task3 spawned as well
        assert len(self.get_task_instance.task_by_name('Task1')) == 2
        assert len(self.get_task_instance.task_by_name('Task2')) == 2
        assert len(self.get_task_instance.task_by_name('Task3')) == 2
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict['finished_nodes']['Task1']) == 2
        assert len(state_dict.get('active_nodes')) == 2

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(self.get_task_instance.task_by_name('Task1')) == 2
        assert len(self.get_task_instance.task_by_name('Task2')) == 2
        assert len(self.get_task_instance.task_by_name('Task3')) == 2
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict['finished_nodes']['Task1']) == 2
        assert len(state_dict.get('active_nodes')) == 2

        # Task3 has finished
        task3_1 = self.get_task('Task3')
        self.set_finished(task3_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(self.get_task_instance.task_by_name('Task1')) == 3
        assert len(self.get_task_instance.task_by_name('Task2')) == 2
        assert len(self.get_task_instance.task_by_name('Task3')) == 2
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict['finished_nodes']['Task1']) == 2
        assert len(state_dict['finished_nodes']['Task2']) == 1
        assert len(state_dict.get('active_nodes')) == 2

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(self.get_task_instance.task_by_name('Task1')) == 3
        assert len(self.get_task_instance.task_by_name('Task2')) == 2
        assert len(self.get_task_instance.task_by_name('Task3')) == 2
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict['finished_nodes']['Task1']) == 2
        assert len(state_dict['finished_nodes']['Task2']) == 1
        assert len(state_dict.get('active_nodes')) == 2

        # Task2 has finished
        task2_1 = self.get_task('Task2')
        self.set_finished(task2_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(self.get_task_instance.task_by_name('Task1')) == 3
        assert len(self.get_task_instance.task_by_name('Task2')) == 2
        assert len(self.get_task_instance.task_by_name('Task3')) == 2
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict['finished_nodes']['Task1']) == 2
        assert len(state_dict['finished_nodes']['Task2']) == 2
        assert len(state_dict.get('active_nodes')) == 1

        # Once more with no change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args == node_args
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(self.get_task_instance.task_by_name('Task1')) == 3
        assert len(self.get_task_instance.task_by_name('Task2')) == 2
        assert len(self.get_task_instance.task_by_name('Task3')) == 2
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict['finished_nodes']['Task1']) == 2
        assert len(state_dict['finished_nodes']['Task2']) == 2
        assert len(state_dict.get('active_nodes')) == 1

    def test_2_to_1_recursive(self):
        #
        # flow1:
        #
        #    Task1      Task2 <-
        #      |          |     |
        #       ----------      |
        #           |           |
        #         Task3   ------
        #
        # Note:
        #   Task1 finishes before Task2 in the first iteration
        #   Task2 finishes before Task1 in the second iteration
        #
        #   Also note that Task3 is run:
        #      With Task1_1 from the first iteration and Task2 from the first iteration
        #      With Task1_1 from the first iteration and Task2 from the second iteration
        #
        edge_table = {
            'flow1': [{'from': ['Task1', 'Task2'], 'to': ['Task3'], 'condition': self.cond_true},
                      {'from': ['Task3'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1', 'Task2'], 'condition': self.cond_true}]
        }
        self.init(edge_table)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 2

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 0
        assert len(state_dict.get('active_nodes')) == 2

        # Task1 has finished
        task1 = self.get_task('Task1')
        self.set_finished(task1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 1

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 1
        assert len(state_dict.get('active_nodes')) == 1

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 2
        assert len(state_dict.get('active_nodes')) == 1

        # No change so far
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert 0 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 2
        assert len(state_dict.get('active_nodes')) == 1

        # Task3 has finished
        task3 = self.get_task('Task3')
        self.set_finished(task3)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict.get('active_nodes')) == 1

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict.get('active_nodes')) == 1

        # Newly spawned Task2 has finished, Task3 should be spawned
        task2_1 = self.get_task('Task2')
        self.set_finished(task2_1)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'Task2' in self.instantiated_tasks
        assert 'Task3' in self.instantiated_tasks
        assert len(self.get_task_instance.task_by_name('Task2')) == 2
        assert len(self.get_task_instance.task_by_name('Task3')) == 2
        assert len(state_dict.get('waiting_edges')) == 2
        assert 0 in state_dict['waiting_edges']
        assert 1 in state_dict['waiting_edges']
        assert len(state_dict.get('finished_nodes')) == 3
        assert len(state_dict['finished_nodes']['Task2']) == 2
        assert len(state_dict.get('active_nodes')) == 1

        task3_1 = self.get_task('Task3')

        # Check parent tasks of Task3 from the first iteration and Task3_1 from the second iteration
        assert task3_1.parent['Task1'] == task3.parent['Task1']
        assert task3_1.parent['Task1'] == task1.task_id

        assert task3_1.parent['Task2'] != task3.parent['Task2']
        assert task3.parent['Task2'] == task2.task_id
        assert task3_1.parent['Task2'] == task2_1.task_id



#!/bin/env python3
# pragma: no cover

from functools import reduce


def fieldExist(message, key):
    try:
        reduce(lambda m, k: m[k], key if isinstance(key, list) else [key], message)
        return True
    except:
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import pytest
import flexmock
from selinon_test_case import SelinonTestCase
from selinon import SystemState
from selinon import Cache
from selinon.errors import DispatcherRetry
from selinon.errors import CacheMissError
from selinon import DataStorage
from celery.result import AsyncResult


class TestCacheIssues(SelinonTestCase):
    class MyStorage(DataStorage):
        def __init__(self):
            self.data = None

        def connect(self):
            # shouldn't be called
            raise NotImplementedError()

        def disconnect(self):
            # called on destruction
            pass

        def is_connected(self):
            return True

        def store(self, node_args, flow_name, task_name, task_id, result):
            self.data = result

        def retrieve(self, flow_name, task_name, task_id):
            return self.data

    class MyCacheAddFailure(Cache):
        def add(self, item_id, item, task_name=None, flow_name=None):
            raise ValueError("Some error when adding to cache")

        def get(self, item_id, task_name=None, flow_name=None):
            raise CacheMissError("Result is not found in the cache")

    class MyCacheGetFailure(Cache):
        def add(self, item_id, item, task_name=None, flow_name=None):
            pass

        def get(self, item_id, task_name=None, flow_name=None):
            raise IndexError("Some error when getting values from cache")

    @pytest.mark.parametrize("cache", (MyCacheAddFailure(), MyCacheGetFailure()))
    def test_async_result_cache_issue(self, cache):
        #
        # flow1:
        #
        #     Task1
        #
        # This test ensures that async result cache issues do not affect flow behaviour.
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true}],
        }
        self.init(edge_table, async_result_cache={'flow1': cache})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None

        task1 = self.get_task('Task1')
        self.set_finished(task1, "some result")

        # Nothing should be raised here even though the cache fails on get(), flow should continue
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()

        assert retry is None

    @pytest.mark.parametrize("cache", (MyCacheAddFailure(), MyCacheGetFailure()))
    def test_result_cache_issue(self, cache):
        #
        # flow1:
        #
        #     Task1
        #
        # Checks that if there is an issue with retrieving task result from result cache in condition, dispatcher
        # will use directly storage.
        edge_table = {
            'flow1': [{'from': [], 'to': ['Task1'], 'condition': self.cond_true},
                      {'from': ['Task1'], 'to': 'Task2', 'condition': lambda db, _: db.get('Task1')}]
        }
        self.init(
            edge_table,
            storage_mapping={'Storage1': self.MyStorage()},
            task2storage_mapping={'Task1': 'Storage1'},
            storage2storage_cache={'Storage1': cache}
        )

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None

        task1 = self.get_task('Task1')
        self.set_finished(task1, "some result")

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

import time
import datetime
from selinon_test_case import SelinonTestCase
from selinon import SystemState


class TestFlow(SelinonTestCase):
    def test_simple_flow(self):
        #
        # flow1:
        #
        #     Task1
        #       |
        #       |
        #     flow2
        #       |
        #       |
        #     Task2
        #
        # Note:
        #    Result of Task1 is not propagated to flow2, parent are not propagated as well
        #
        edge_table = {
            'flow1': [{'from': ['Task1'], 'to': ['flow2'], 'condition': self.cond_true},
                      {'from': ['flow2'], 'to': ['Task2'], 'condition': self.cond_true},
                      {'from': [], 'to': ['Task1'], 'condition': self.cond_true}],
            'flow2': []
        }
        queues = {
            'Task1': 'mytask1queue',
            'Task2': 'mytask2queue',
        }
        flow2_queue = 'flow2_queue'
        dispatcher_queues = {
            'flow2': flow2_queue,
            'flow1': 'flow1_queue'
        }
        self.init(edge_table, task_queues=queues, dispatcher_queues=dispatcher_queues)

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'flow2' not in self.instantiated_flows
        assert 'Task2' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # check task's queue configuration
        task1 = self.get_task('Task1')
        assert task1.queue == queues.get(task1.task_name)

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'Task1' in self.instantiated_tasks
        assert 'flow2' not in self.instantiated_flows
        assert 'Task2' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0

        # Task1 has finished
        task1_result = [1, 2, 3, 4]
        self.set_finished(task1, task1_result)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        flow2 = self.get_flow('flow2')

        assert retry is not None
        assert system_state.node_args is None
        assert 'flow2' in self.instantiated_flows
        assert flow2.node_args is None
        assert flow2.flow_name == 'flow2'
        assert flow2.parent is None
        assert 'Task1' in self.instantiated_tasks
        assert 'flow2' in self.instantiated_flows
        assert 'Task2' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0
        assert len(state_dict['finished_nodes']) == 1
        assert len(state_dict['active_nodes']) == 1

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'flow2' in self.instantiated_flows
        assert flow2.node_args is None
        assert flow2.flow_name == 'flow2'
        assert flow2.parent is None
        assert 'Task1' in self.instantiated_tasks
        assert 'flow2' in self.instantiated_flows
        assert 'Task2' not in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 1
        assert state_dict['waiting_edges'][0] == 0
        assert len(state_dict['finished_nodes']) == 1
        assert len(state_dict['active_nodes']) == 1

        # check flow queue propagation
        flow2 = self.get_flow('flow2')
        assert flow2.queue == flow2_queue

        # flow2 has finished
        self.set_finished(flow2, {'TaskSubflow': [1]})

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        flow2 = self.get_flow('flow2')

        assert retry is not None
        assert system_state.node_args is None
        assert 'flow2' in self.instantiated_flows
        assert flow2.node_args is None
        assert flow2.flow_name == 'flow2'
        assert flow2.parent is None
        assert 'Task1' in self.instantiated_tasks
        assert 'flow2' in self.instantiated_flows
        assert 'Task2' in self.instantiated_tasks
        assert flow2.node_args is None
        assert flow2.parent is None
        assert len(state_dict.get('waiting_edges')) == 2
        assert state_dict['waiting_edges'][0] == 0
        assert len(state_dict['finished_nodes']) == 2
        assert len(state_dict['active_nodes']) == 1

        # No change
        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is not None
        assert system_state.node_args is None
        assert 'flow2' in self.instantiated_flows
        assert 'Task1' in self.instantiated_tasks
        assert 'flow2' in self.instantiated_flows
        assert 'Task2' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert state_dict['waiting_edges'][0] == 0
        assert len(state_dict['finished_nodes']) == 2
        assert len(state_dict['active_nodes']) == 1

        # Task2 has finished
        task2 = self.get_task('Task2')
        self.set_finished(task2)

        system_state = SystemState(id(self), 'flow1', state=state_dict, node_args=system_state.node_args)
        retry = system_state.update()
        state_dict = system_state.to_dict()

        assert retry is None
        assert system_state.node_args is None
        assert task2.node_args is None
        assert 'flow2' in self.instantiated_flows
        assert 'Task1' in self.instantiated_tasks
        assert 'flow2' in self.instantiated_flows
        assert 'Task2' in self.instantiated_tasks
        assert len(state_dict.get('waiting_edges')) == 2
        assert state_dict['waiting_edges'][0] == 0
        assert len(state_dict['finished_nodes']) == 3
        assert len(state_dict['active_nodes']) == 0

    def test_throttle(self):
        #
        # flow1:
        #           |
        #         flow2
        #
        # Note:
        #    flow2 should be throttled by 2s in next flow1 run
        #
        edge_table = {
            'flow1': [{'from': [], 'to': ['flow2'], 'condition': self.cond_true}],
            'flow2': []
        }
        self.init(edge_table, throttle_flows={'flow2': datetime.timedelta(seconds=2)})

        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()

        assert retry is not None
        assert 'flow2' in self.instantiated_flows
        assert self.get_flow('flow2').countdown is None

        # Let's sleep to ensure we get less then 2s delay
        time.sleep(0.01)

        # run flow for the second time, we should be postponed by ~2s
        system_state = SystemState(id(self), 'flow1')
        retry = system_state.update()

        assert retry is not None
        assert self.get_flow('flow2').countdown is not None
        assert self.get_flow('flow2').countdown < 2





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################

from selinon import DataStorage


class MySimpleStorage(DataStorage):
    def __init__(self, connection_string):
        super().__init__()
        self.connection_string = connection_string

    def is_connected(self):
        pass

    def connect(self):
        pass

    def disconnect(self):
        pass

    def retrieve(self, flow_name, task_name, task_id):
        pass

    def store(self, node_args, flow_name, task_name, task_id, result):
        pass

    def trace(self, event, msg_dict):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ######################################################################
# Copyright (C) 2016-2018  Fridolin Pokorny, fridolin.pokorny@gmail.com
# This file is part of Selinon project.
# ######################################################################
"""Abstract representation of nodes in task/flow dependencies - a node is either a task or a flow."""

import abc
import datetime
import os
import re

from .errors import ConfigurationError
from .global_config import GlobalConfig


class Node(metaclass=abc.ABCMeta):
    """An abstract class for node representation."""

    _NAME_RE = re.compile(r"^[_a-zA-Z][_a-zA-Z0-9]*$")

    def __init__(self, name):
        """Instantiate a node (flow/task)."""
        if not self.check_name(name):
            raise ConfigurationError("Invalid node name '%s'" % name)
        self._name = name

    @property
    def name(self):
        """Get name of the node.

        :return: a name of the node
        """
        return self._name

    def _expand_queue_name(self, queue_name):
        """Assign queue name based on configuration, do expansion based on env variables if needed.

        :param queue_name: name of queue as provided in config file or None
        :return: final name of the queue
        """
        if not queue_name:
            queue_name = GlobalConfig.default_dispatcher_queue if self.is_flow() else GlobalConfig.default_task_queue
        try:
            return queue_name.format(**os.environ)
        except KeyError as exc:
            if self.is_flow():
                err_msg = "Expansion of queue name based on environment variables failed for flow '%s', queue: '%s'" \
                          % (self.name, queue_name)
            else:
                err_msg = "Expansion of queue name based on environment variables failed for task '%s', queue: '%s' " \
                          % (self.name, queue_name)
            raise ConfigurationError(err_msg) from exc

    def is_flow(self):
        """Check if this node is a flow.

        :return: True if node represents a Flow
        """
        from .flow import Flow
        return isinstance(self, Flow)

    def is_task(self):
        """Check if this node is a task.

        :return: True if node represents a Task
        """
        from .task import Task
        return isinstance(self, Task)

    @classmethod
    def check_name(cls, name):
        """Check whether name is a correct node (flow/task) name.

        :param name: node name
        :return: True if name is a correct node name
        :rtype: bool
        """
        return cls._NAME_RE.match(name)

    def parse_throttling(self, dict_):
        """Parse throttling definition from a dictionary.

        :param dict_: dictionary from which throttling should be parsed
        :return: timedelta describing throttling countdown or None if not throttling applied
        :rtype: time.timedelta
        """
        if not dict_:
            return None

        if not isinstance(dict_, dict):
            raise ConfigurationError("Definition of throttling expects key value definition, got %s instead in '%s'"
                                     % (dict_['throttling'], self.name))
        try:
            return datetime.timedelta(**dict_)
        except TypeError as exc:
            raise ConfigurationError("Wrong throttling definition in '%s', expected values are %s"
                                     % (self.name, ['days', 'seconds', 'microseconds', 'milliseconds', 'minutes',
                                                    'hours', 'weeks'])) from exc
